/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19 janvier 2007
   derniere modification le 20/03/2007 */

#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECore.h>
#include <SCE/interface/SCEScene.h>

int SCE_Init_Scene (void)
{
    return SCE_OK;
}
void SCE_Quit_Scene (void)
{
}


/* ajoute le 23/02/2008 */
void SCE_Scene_Init (SCE_SScene *scene)
{
    scene->rootnode = NULL;
    scene->models = NULL;
    scene->lights = NULL;
    scene->cameras = NULL;
    scene->frustum_culling = SCE_FALSE;
}

/* ajoute le 23/02/2008 */
/* revise le 20/03/2008 */
SCE_SScene* SCE_Scene_Create (void)
{
    SCE_SCamera *cam = NULL;
    SCE_SScene *scene = NULL;
    SCE_btstart ();
    if (!(scene = SCE_malloc (sizeof *scene)))
        Logger_LogSrc ();
    else
    {
        SCE_Scene_Init (scene);
#define SCE_SCENE_ASSERT(c)\
        if (c)\
        {\
            SCE_Scene_Delete (scene);\
            Logger_LogSrc ();\
            SCE_btend ();\
            return NULL;\
        }

        SCE_SCENE_ASSERT (!(scene->models = SCE_List_Create (NULL)))
        SCE_SCENE_ASSERT (!(scene->lights = SCE_List_Create (NULL)))
        SCE_SCENE_ASSERT (!(scene->cameras = SCE_List_Create (NULL)))

        SCE_SCENE_ASSERT (!(cam = SCE_Camera_Create ()))
        SCE_SCENE_ASSERT (SCE_Scene_AddCamera (scene, cam) < 0)
        if (!(scene->rootnode = SCE_Node_Create ()))
        {
            SCE_Scene_Delete (scene), scene = NULL;
            Logger_LogSrc ();
        }
    }
    SCE_btend ();
    return scene;
}
/* ajoute le 23/02/2008 */
void SCE_Scene_Delete (SCE_SScene *scene)
{
    if (scene)
    {
        SCE_Node_Delete (scene->rootnode);
        SCE_List_Delete (scene->cameras);
        SCE_List_Delete (scene->lights);
        SCE_List_Delete (scene->models);
        SCE_free (scene);
    }
}


/* ajoute le 20/03/2008 */
void SCE_Scene_EnableFrustumCulling (SCE_SScene *scene)
{
    scene->frustum_culling = SCE_TRUE;
    SCE_Scene_GenerateBoundingsBoxes (scene); /* TODO: bof... ? */
}
/* ajoute le 20/03/2008 */
void SCE_Scene_DisableFrustumCulling (SCE_SScene *scene)
{
    scene->frustum_culling = SCE_FALSE;
}


/* ajoute le 20/03/2008 */
void SCE_Scene_GenerateBoundingsBoxes (SCE_SScene *scene)
{
    SCE_SListIterator *it = NULL;
    SCE_List_ForEach (it, scene->models)
        SCE_Model_GenerateBoundingBox (SCE_List_GetData (it));
}


/* ajoute le 26/02/2008 */
SCE_SNode* SCE_Scene_GetRootNode (SCE_SScene *scene)
{
    return scene->rootnode;
}

/* ajoute le 20/03/2008 */
SCE_SCamera* SCE_Scene_GetCamera (SCE_SScene *scene, unsigned int index)
{
    SCE_SListIterator *it = NULL;
    it = SCE_List_GetIterator (scene->cameras, index);
    return SCE_List_GetData (it);
}

/* ajoute le 26/02/2008 */
int SCE_Scene_AddNode (SCE_SScene *scene, SCE_SNode *node)
{
    if (SCE_Node_Attach (scene->rootnode, node) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* ajoute le 26/02/2008 */
/* revise le 10/03/2008 */
int SCE_Scene_AddModel (SCE_SScene *scene, SCE_SModel *model)
{
    if (SCE_List_AddBack (scene->models, model) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* ajoute le 20/03/2008 */
int SCE_Scene_AddLight (SCE_SScene *scene, SCE_SLight *light)
{
    if (SCE_List_AddBack (scene->lights, light) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}

/* ajoute le 20/03/2008 */
int SCE_Scene_AddCamera (SCE_SScene *scene, SCE_SCamera *cam)
{
    if (SCE_List_AddBack (scene->cameras, cam) < 0)
    {
        Logger_LogSrc ();
        return SCE_ERROR;
    }
    return SCE_OK;
}


/* ajoute le 19/03/2008 */
/* revise le 23/03/2008 */
void SCE_Scene_Render (SCE_SScene *scene)
{
    SCE_SListIterator *i = NULL, *j = NULL, *k = NULL;
    SCE_SModel *mdl = NULL;
    SCE_SLight *light = NULL;
    SCE_SCamera *cam = NULL;
    int render;
    SCE_TVector3 vec1, vec2;
    SCE_SBoundingBox *box = NULL;
    float length;
    float *p = NULL;

    SCE_CClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /* pour chaque camera active */
    SCE_List_ForEach (i, scene->cameras)
    {
        cam = SCE_List_GetData (i);

        /* activation de la camera et calcul du frustum (si necessaire) */
        SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
        SCE_CPushMatrix ();
        SCE_Camera_Use (cam);
        if (scene->frustum_culling)
            SCE_Camera_UpdateFrustum (cam);

        /* assignation de la matrice camera au gestionnaire de lumieres */
        SCE_Light_SetCameraMatrix (SCE_Camera_GetView (cam));

        /* pour chaque modele de la scene */
        SCE_List_ForEach (j, scene->models)
        {
            mdl = SCE_List_GetData (j);

            /* on verifie si le modele est dans le frustum */
            box = SCE_Model_GetBoundingBox (mdl);
            if (scene->frustum_culling)
                render = SCE_Frustum_BoundingBoxIn (SCE_Camera_GetFrustum (cam),
                                                    box);
            else
                render = SCE_TRUE;
            if (!render)
                continue;

            /* obtention de donnees necessaire au calcul des lumieres */
            SCE_BoundingBox_GetCenterv (box, vec1);
            SCE_Vector3_Copy (vec2, &(SCE_BoundingBox_GetPoints (box)[0]));
            length = SCE_Vector3_Distance (vec1, vec2);

            /* activation des lumieres */
            SCE_List_ForEach (k, scene->lights)
            {
                light = SCE_List_GetData (k);

                /* on n'active la lumiere que si le modele est a sa portee */
                p = SCE_Light_GetPosition (light);
                if (SCE_Vector3_Distance (p, vec1) <=
                    SCE_Light_GetRadius (light) + length+3.0) /* TODO: shit hack */
                    SCE_Light_Use (light);
            }

            /* rendu du modele */
            SCE_Model_Render (mdl);
            SCE_Light_Use (NULL);
        }

        SCE_CSetActiveMatrix (SCE_MAT_MODELVIEW);
        SCE_CPushMatrix ();
    }
}
